#include <stdio.h>
#include <stdlib.h>
#include <qdos.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include "wild.h"
#include "xprdefs.h"

#define case_map(c) toupper(c)

static int recmatch (char *, char *);

static char *isshexp (char *p)	/* candidate sh expression */
/* If p is a sh expression, a pointer to the first special character is
   returned.  Otherwise, NULL is returned. */
{
    for (; *p; p++)
	if (*p == '\\' && *(p + 1))
	    p++;
#ifdef REGEXP
	else if (*p == '?' || *p == '*' || *p == '[')
#else
	else if (*p == '?' || *p == '*' || *p == '!')
#endif
	    return p;
    return NULL;
}


static int recmatch (char *p, char *s)
	       /* sh pattern to match */
		/* string to match it to */
/* Recursively compare the sh pattern p with the string s and return 1 if
   they match, and 0 or 2 if they don't or if there is a syntax error in the
   pattern.  This routine recurses on itself no deeper than the number of
   characters in the pattern. */
{
    int c;			/* pattern char or start of range in [-] loop */

    /* Get first character, the pattern for new recmatch calls follows */
    c = *p++;

    /* If that was the end of the pattern, match if string empty too */
    if (c == 0)
	return *s == 0;

    /* '?' (or '%') matches any character (but not an empty string) */
#ifdef REGEXP
    if (c == '?')
#else
    if (c == '?' || c == '%')
#endif
	return *s ? recmatch (p, s + 1) : 0;

    /* '*' matches any number of characters, including zero */
    if (c == '*')
    {
	if (*p == 0)
	    return 1;
	for (; *s; s++)
	    if ((c = recmatch (p, s)) != 0)
		return c;
	return 2;		/* 2 means give up--shmatch will return false */
    }
#ifdef REGEXP
    /* Parse and process the list of characters and ranges in brackets */
    if (c == '[')
    {
	int e;			/* flag true if next char to be taken literally */
	char *q;		/* pointer to end of [-] group */
	int r;			/* flag true to match anything but the range */

	if (*s == 0)		/* need a character to match */
	    return 0;
	p += (r = *p == '!');	/* see if reverse */
	for (q = p, e = 0; *q; q++)	/* find closing bracket */
	    if (e)
		e = 0;
	    else if (*q == '\\')
		e = 1;
	    else if (*q == ']')
		break;
	if (*q != ']')		/* nothing matches if bad syntax */
	    return 0;
	for (c = 0, e = *p == '-'; p < q; p++)	/* go through the list */
	{
	    if (e == 0 && *p == '\\')	/* set escape flag if \ */
		e = 1;
	    else if (e == 0 && *p == '-')	/* set start of range if - */
		c = *(p - 1);
	    else
	    {
		if (*(p + 1) != '-')
		    for (c = c ? c : *p; c <= *p; c++)	/* compare range */
			if (case_map (c) == case_map (*s))
			    return r ? 0 : recmatch (q + 1, s + 1);
		c = e = 0;	/* clear range, escape flags */
	    }
	}
	return r ? recmatch (q + 1, s + 1) : 0;		/* bracket match failed */
    }
#endif
    /* If escape ('\'), just compare next character */
    if (c == '\\')
	if ((c = *p++) == 0)	/* if \ at end, then syntax error */
	    return 0;

    /* Just a character--compare it */
    return case_map (c) == case_map (*s) ? recmatch (p, ++s) : 0;
}


static char *stpcpy (char *d, const char *s)
{
    while (*d++ = *s++) ;
    return (d - 1);
}

int ScanDir (char *dnam, FLIST ** fl, short recurse)
{
    static char match[40] = {0};
    static char ddev[8] =  {0};
    static short l = 0;
    static int nc = 0;
    int chid;
    struct qdirect qd;
    char *dp;

    if (l == 0)
    {
        nc = 0;
	if (isdirdev (dnam))
	{
	    dp = dnam;
	    strncpy (ddev, dnam, 5);
	}
	else
	{
	    char temp[40];
	    char *p;

	    getcwd (temp, 40);
	    p = (temp + strlen (temp) - 1);
	    if (*p != '_')
	    {
		*p++ = '_';
		*p = 0;
	    }
	    strncpy (ddev, temp, 5);
	    dp = alloca ((p - temp) + 8);
	    p = stpcpy (dp, temp);
	    strcpy (p, dnam);
	}
	{
	    char *q = strpbrk (dp, "*?");

	    strcpy (match, dp + 5);
	    if (q)
	    {
		while (q != dp && *q != '_')
		{
		    q--;
		}
		*(++q) = '\0';
	    }
	}
    }
    else
    {
	dp = dnam;
    }
    if ((chid = io_open (dp, 4)) > 0)
    {
	while (io_fstrg (chid, -1, &qd, 64) > 0)
	{
	    short j;

	    if (qd.d_szname)
	    {
		if (qd.d_type == 255)
		{
                    if(recurse)
                    {
		        char fnam[40], *p;

		        p = stpcpy (fnam, ddev);
		        strncpy (p, qd.d_name, qd.d_szname);
		        *(p + qd.d_szname) = 0;
		        l++;
		        (void) ScanDir (fnam, fl, recurse);
		        l--;
		    }
		    else
		    {
		        continue;
		    }
		}
		else
		{
		    char nam[40];
		    strncpy (nam, qd.d_name, qd.d_szname);
		    *(nam + qd.d_szname) = 0;
		    if (recmatch (match, nam) == 1)
		    {
                        char txt[48];
		    	FLIST *pl;
                
			strcpy (txt, ddev);
			strcpy (txt + 5, nam);
                        for(pl = *fl;pl;pl = pl->next)
                        {
                            if(stricmp(txt, pl->text) == 0)
                            {
                                break;
                            }
                        }
                        if(pl == NULL)
                        {
                            AddFile(fl, txt);
                            nc++;
                        }
		    }
		}
	    }
	}
	io_close (chid);
    }
    if (l == 0)
    {
	*ddev = 0;
	*match = 0;
    }
    return nc;
}
void AddFile(FLIST ** fl, char *txt)
{
    FLIST *pl;
    pl = malloc (sizeof (FLIST) + strlen(txt)+4);
    strcpy(pl->text, txt);
    pl->next = NULL;
    ut_link (fl, pl);
}
void FreeWildList(FLIST **fl)
{
    FLIST *l, *ll;
    for(l = *fl; l;)
    {
        ll = l;
        l = l->next;
        ut_unlnk(fl, ll);
        free(ll);
    }
    *fl = NULL;
}
#ifdef TEST
main (int ac, char **av)
{
    FLIST *h = NULL, *l;

    if (av[1])
    {
	ScanDir (av[1], &h);
        for (l = h; l; l = l->next)
        {
            puts(l->text);
	}
    }
}
#endif
